'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var FS = require('fs');
var FS__default = _interopDefault(FS);
var Path = require('path');
var Path__default = _interopDefault(Path);
var vueVirtualTextdocument = require('@vuedx/vue-virtual-textdocument');
var vscodeUri = require('vscode-uri');
var analyze = require('@vuedx/analyze');
var QuickLRU = _interopDefault(require('quick-lru'));
var templateAstTypes = require('@vuedx/template-ast-types');
var deIndent = _interopDefault(require('de-indent'));

const PATCHED_METHODS = Symbol('Vue Patched Methods');
function tryPatchMethod(target, methodName, createOverride) {
    const patched = target[PATCHED_METHODS] || [];
    if (patched.includes(methodName))
        return;
    let fn = target[methodName];
    if (typeof fn === 'function') {
        try {
            fn = fn.bind(target);
        }
        catch (_a) {
            // - ignore
        }
    }
    target[methodName] = createOverride(fn);
    patched.push(methodName);
    target[PATCHED_METHODS] = patched;
}

function getLastNumberFromVersion(version) {
    const parts = version.split(/[^0-9]+/);
    const ver = parts.pop();
    return Number(ver);
}
class ProxyDocumentStore extends vueVirtualTextdocument.DocumentStore {
    get(fileNameOrUri) {
        return super.get(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
    has(fileNameOrUri) {
        return super.has(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
    set(fileNameOrUri, document) {
        return super.set(vueVirtualTextdocument.asUri(fileNameOrUri), document);
    }
    delete(fileNameOrUri) {
        return super.delete(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
}
function getConfig(config = {}) {
    return {
        ...config,
        features: {
            diagnostics: ['semantic', 'suggestion', 'suggestion'],
            organizeImports: true,
            quickInfo: true,
            rename: true,
            refactor: true,
            ...config.features,
        },
        directories: [{ kind: 'component', name: 'components', path: 'src/components' }],
    };
}
class PluginContext {
    constructor(typescript) {
        this.typescript = typescript;
        this._config = getConfig();
        this.moduleResolutionHistory = new Map();
        this.watchers = new Map();
        this.store = new ProxyDocumentStore((uri) => {
            const fileName = vscodeUri.URI.parse(uri).fsPath;
            const content = this.typescript.sys.readFile(fileName) || '';
            return vueVirtualTextdocument.VueTextDocument.create(uri, 'vue', 0, content);
        }, (uri) => {
            const fileName = vscodeUri.URI.parse(uri).fsPath;
            return vscodeUri.URI.file(this.projectService.toPath(fileName)).toString();
        });
    }
    get config() {
        return this._config;
    }
    get serviceHost() {
        return this._serverHost;
    }
    get projectService() {
        return this._projectService;
    }
    log(message) {
        if (this.projectService) {
            this.projectService.logger.info(`Vue.js:: ${message}`);
        }
    }
    getVueVersion(fileName) {
        return '3.0.0';
    }
    createVueDocument(fileName, content) {
        const uri = vscodeUri.URI.file(fileName).toString();
        const document = vueVirtualTextdocument.VueTextDocument.create(uri, 'vue', 0, content);
        this.store.set(uri, document);
        return document;
    }
    error(message) {
        if (this.projectService) {
            this.projectService.logger.msg(`Vue.js:: ${message} ${message.stack}`, this.typescript.server.Msg.Err);
        }
    }
    getWatchers(fileName) {
        const containingFile = vueVirtualTextdocument.isVirtualFile(fileName) ? vueVirtualTextdocument.getContainingFile(fileName) : fileName;
        if (!this.watchers.has(containingFile)) {
            this.watchers.set(containingFile, new Map());
        }
        return this.watchers.get(containingFile);
    }
    watchVirtualFile(fileName, callback) {
        const watchers = this.getWatchers(fileName);
        if (!watchers.has(fileName)) {
            watchers.set(fileName, new Set());
        }
        watchers.get(fileName).add(callback);
    }
    triggerVirtualFileWatchers(fileName, baseEvent) {
        var _a;
        const watchers = this.getWatchers(fileName);
        if (vueVirtualTextdocument.isVirtualFile(fileName)) {
            (_a = watchers.get(fileName)) === null || _a === void 0 ? void 0 : _a.forEach((callback) => {
                callback(fileName, baseEvent);
            });
        }
        else {
            watchers.forEach((watchers) => {
                watchers.forEach((callback) => {
                    callback(fileName, baseEvent);
                });
            });
        }
    }
    stopVirtualFileWatcher(fileName, callback) {
        const watchers = this.getWatchers(fileName).get(fileName);
        if (watchers) {
            watchers.delete(callback);
        }
    }
    tryCreateScriptInfo(fileName) {
        try {
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                this.store.get(vueVirtualTextdocument.getContainingFile(fileName));
                const scriptInfo = this.projectService.getOrCreateScriptInfoForNormalizedPath(this.typescript.server.toNormalizedPath(vueVirtualTextdocument.getContainingFile(fileName)), false);
                if (scriptInfo) {
                    patchScriptInfo(this, scriptInfo);
                }
            }
        }
        catch (_a) { }
    }
    load(info) {
        this.log(`Loading Vue plugin: ${info.project.getProjectName()}`);
        this._serverHost = info.serverHost;
        this._projectService = info.project.projectService;
        this.setConfig(info.config);
        patchProjectService(this);
        patchServiceHost(this);
        patchLanguageServiceHost(this, info.languageServiceHost);
    }
    setConfig(config) {
        this._config = getConfig(config);
    }
}
function patchProjectService(context) {
    patchExtraFileExtensions(context);
}
function patchExtraFileExtensions(context) {
    var _a;
    const extraFileExtensions = [
        { extension: 'vue', isMixedContent: false, scriptKind: context.typescript.ScriptKind.Deferred },
    ];
    tryPatchMethod(context.projectService, 'setHostConfiguration', (setHostConfiguration) => {
        context.log(`[patch] Add support for vue extension. (ProjectService)`);
        return (args) => {
            const current = context.projectService.hostConfiguration
                .extraFileExtensions;
            if (args.extraFileExtensions) {
                args.extraFileExtensions.push(...extraFileExtensions);
                context.log(`extraFileExtensions: ${JSON.stringify(args.extraFileExtensions)}`);
            }
            else if (!current || !current.some((ext) => ext.extension === 'vue')) {
                args.extraFileExtensions = [...extraFileExtensions];
            }
            return setHostConfiguration(args);
        };
    });
    if ((_a = context.projectService.hostConfiguration.extraFileExtensions) === null || _a === void 0 ? void 0 : _a.some((ext) => ext.extension === 'vue')) {
        return;
    }
    // Enable .vue after enhancing the language server.
    context.projectService.setHostConfiguration({ extraFileExtensions: [] });
}
function patchLanguageServiceHost(context, languageServiceHost) {
    patchGetScriptFileNames(context, languageServiceHost);
    patchModuleResolution(context, languageServiceHost);
    tryPatchMethod(languageServiceHost, 'getCompilationSettings', (getCompilationSettings) => () => {
        const settings = getCompilationSettings();
        settings.jsx = context.typescript.JsxEmit.Preserve;
        return settings;
    });
}
function patchGetScriptFileNames(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'getScriptFileNames', (getScriptFileNames) => {
        context.log(`[patch] Override getScriptFileNames to expand .vue files to virtual files. (LanguageServerHost)`);
        const previousVueFiles = new Set();
        return () => {
            const fileNames = new Set();
            [...getScriptFileNames(), ...previousVueFiles]
                .map((fileName) => (vueVirtualTextdocument.isVirtualFile(fileName) ? vueVirtualTextdocument.getContainingFile(fileName) : fileName))
                .forEach((fileName) => {
                if (vueVirtualTextdocument.isVueFile(fileName)) {
                    const document = context.store.get(fileName);
                    if (document) {
                        previousVueFiles.add(fileName);
                        fileNames.add(document.getDocumentFileName(vueVirtualTextdocument.MODULE_SELECTOR));
                    }
                    else {
                        previousVueFiles.delete(fileName);
                    }
                }
                else {
                    fileNames.add(fileName);
                }
            });
            return Array.from(fileNames);
        };
    });
}
function patchFileExists(context) {
    tryPatchMethod(context.serviceHost, 'fileExists', (fileExists = context.typescript.sys.fileExists) => {
        context.log(`[patch] Override fileExists to check containing file for virtual files. (ServiceHost)`);
        return (fileName) => {
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                const document = context.store.get(vueVirtualTextdocument.getContainingFile(fileName));
                const { selector } = vueVirtualTextdocument.parseVirtualFileName(fileName);
                return fileName === (document === null || document === void 0 ? void 0 : document.getDocumentFileName(selector));
            }
            return fileExists(fileName);
        };
    });
}
function patchReadFile(context) {
    tryPatchMethod(context.serviceHost, 'readFile', (readFile) => {
        context.log(`[patch] Override readFile to check containing file for virtual files. (ServiceHost)`);
        return (fileName) => {
            var _a;
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                context.log(`host.readFile("${fileName}")`);
                const document = (_a = context.store.get(vueVirtualTextdocument.getContainingFile(fileName))) === null || _a === void 0 ? void 0 : _a.getDocument(fileName);
                if (document)
                    return document.getText();
                return;
            }
            return readFile ? readFile(fileName) : context.typescript.sys.readFile(fileName);
        };
    });
}
function patchModuleResolution(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'resolveModuleNames', (resolveModuleNames) => {
        context.log(`[patch] Override resolveModuleNames to resolve imports from .vue files. (LanguageServerHost)`);
        return (moduleNames, containingFile, reusedNames, redirectedReferences, options) => {
            if (vueVirtualTextdocument.isVueFile(containingFile))
                throw new Error('A .vue file should not be part of TS program.');
            const newModuleNames = moduleNames.map((moduleName) => vueVirtualTextdocument.isVueFile(moduleName)
                ? moduleName + vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR + vueVirtualTextdocument.MODULE_SELECTOR
                : moduleName.endsWith('.vue?internal')
                    ? moduleName.replace(/\?internal$/, '') + vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR + vueVirtualTextdocument.INTERNAL_MODULE_SELECTOR
                    : moduleName);
            // TODO: Support paths mapped to .vue files, if needed.
            const result = resolveModuleNames
                ? resolveModuleNames(newModuleNames, containingFile, reusedNames, redirectedReferences, options)
                : [];
            const history = new Map();
            result.map((resolved, index) => {
                if (resolved && vueVirtualTextdocument.isVirtualFile(resolved.resolvedFileName)) {
                    history.set(moduleNames[index], vueVirtualTextdocument.getContainingFile(resolved.resolvedFileName));
                    context.tryCreateScriptInfo(resolved.resolvedFileName); // Load vue file now to avoid filename case insensitivity issues.
                }
            });
            if (history.size &&
                vueVirtualTextdocument.isVirtualFile(containingFile) &&
                containingFile.endsWith('.vue' + vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR + '_render.tsx')) {
                context.moduleResolutionHistory.set(containingFile, history);
            }
            return result;
        };
    });
}
function patchServiceHost(context) {
    patchWatchFile(context);
    patchFileExists(context);
    patchReadFile(context);
}
function patchWatchFile(context) {
    tryPatchMethod(context.serviceHost, 'watchFile', (watchFile) => {
        context.log(`[patch] Override watchFile to watch virtual files. (ServiceHost)`);
        return (fileName, callback, pollingInterval, options) => {
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                context.watchVirtualFile(fileName, callback);
                return {
                    close() {
                        context.stopVirtualFileWatcher(fileName, callback);
                    },
                };
            }
            return watchFile(fileName, callback, pollingInterval, options);
        };
    });
}
function patchScriptInfo(context, scriptInfo) {
    if (!scriptInfo)
        throw new Error('ScriptInfo is required.');
    tryPatchMethod(scriptInfo, 'editContent', (editContent) => {
        context.log(`[patch] Override editContent() of "${scriptInfo.fileName}" to sync virtual files. (ScriptInfo)`);
        return (start, end, newText) => {
            const document = context.store.get(scriptInfo.fileName);
            if (!document)
                throw new Error('VueTextDocument should exist for every ScriptInfo.');
            const range = { start: document.positionAt(start), end: document.positionAt(end) };
            editContent(start, end, newText);
            context.log(`UPDATE ${scriptInfo.getLatestVersion()} ==> ${scriptInfo.fileName} ${start}:${end} ${newText}`);
            vueVirtualTextdocument.VueTextDocument.update(document, [{ range, text: newText }], getLastNumberFromVersion(scriptInfo.getLatestVersion()));
            context.triggerVirtualFileWatchers(scriptInfo.fileName, context.typescript.FileWatcherEventKind.Changed);
        };
    });
}

function findTemplateNodeAt(ast, position) {
    return findTemplateNodeFor(ast, position, position);
}
function findTemplateElementNodeAt(ast, position) {
    const result = findTemplateNodeFor(ast, position, position);
    while (result.ancestors.length) {
        if (templateAstTypes.isRootNode(result.node) ||
            templateAstTypes.isElementNode(result.node) ||
            templateAstTypes.isTextNode(result.node) ||
            templateAstTypes.isInterpolationNode(result.node) ||
            templateAstTypes.isCommentNode(result.node)) {
            break;
        }
        result.node = result.ancestors.pop().node;
    }
    return result;
}
function findTemplateNodeFor(ast, start, end) {
    const found = {
        node: null,
        ancestors: [],
    };
    templateAstTypes.traverseEvery(ast, (node, ancestors) => {
        if (node.loc.start.offset <= start && end <= node.loc.end.offset) {
            found.node = node;
            found.ancestors = ancestors.slice();
            return true;
        }
        return false;
    });
    return found;
}
function findTemplateNodesIn(ast, start, end) {
    if (start === end) {
        const a = findTemplateElementNodeAt(ast, start);
        return a.node ? [a.node] : [];
    }
    const a = findTemplateElementNodeAt(ast, start);
    const b = findTemplateElementNodeAt(ast, end);
    if (!a.node || !b.node)
        return [];
    console.log('MATCH IN TEMPLATE ?>>>>>' + JSON.stringify({ a: a.node.loc.source, b: b.node.loc.source }));
    if (a.node === b.node)
        return [a.node];
    if (!a.ancestors.length || !b.ancestors.length)
        return [];
    const pa = a.ancestors.pop().node;
    const pb = b.ancestors.pop().node;
    if (pa === b.node)
        return [pa];
    if (pb === a.node)
        return [pb];
    if (pa === pb) {
        return pa.children.slice(pa.children.indexOf(a.node), 1 + pa.children.indexOf(b.node));
    }
    return [];
}

function createCachedAnalyzer() {
    const cache = new QuickLRU({ maxSize: 1000 });
    const analyzer = analyze.createFullAnalyzer([], { babel: { plugins: ['typescript', 'jsx'] } });
    return (document) => {
        const key = `${document.version}::${document.fsPath}`;
        if (cache.has(key))
            return cache.get(key);
        const info = analyzer.analyze(document.getText(), document.fsPath);
        cache.set(key, info);
        return info;
    };
}
function createServerHelper(context) {
    const getComponentInfo = createCachedAnalyzer();
    function findNodeAtPosition(fileName, position) {
        const document = getRenderDoc(fileName);
        if (document)
            return { document, ...findTemplateNodeAt(document.ast, position) };
        else
            return { node: null, ancestors: [], document };
    }
    function getDocument(fileName) {
        var _a;
        return vueVirtualTextdocument.isVirtualFile(fileName)
            ? (_a = context.store.get(vueVirtualTextdocument.getContainingFile(fileName))) === null || _a === void 0 ? void 0 : _a.getDocument(fileName) : context.store.get(fileName);
    }
    function getVueDocument(fileName) {
        return vueVirtualTextdocument.isVueFile(fileName)
            ? context.store.get(fileName)
            : vueVirtualTextdocument.isVirtualFile(fileName)
                ? context.store.get(vueVirtualTextdocument.getContainingFile(fileName))
                : null;
    }
    function getDocumentAt(fileName, position) {
        const document = context.store.get(fileName); // The file should exist.
        const block = document.blockAt(position);
        if (!block)
            return;
        if (block.type === 'template') {
            return document.getDocument(vueVirtualTextdocument.RENDER_SELECTOR);
        }
        return document.getDocument(document.getBlockSelector(block));
    }
    function isRenderFunctionDocument(document) {
        var _a;
        return ((_a = document) === null || _a === void 0 ? void 0 : _a.selector.type) === vueVirtualTextdocument.RENDER_SELECTOR;
    }
    function isRenderFunctionFileName(fileName) {
        var _a;
        return vueVirtualTextdocument.isVirtualFile(fileName) && ((_a = vueVirtualTextdocument.parseVirtualFileName(fileName)) === null || _a === void 0 ? void 0 : _a.selector.type) === vueVirtualTextdocument.RENDER_SELECTOR;
    }
    function getRenderDoc(fileName) {
        var _a, _b;
        if (vueVirtualTextdocument.isVirtualFile(fileName))
            return (_a = getVueDocument(fileName)) === null || _a === void 0 ? void 0 : _a.getDocument('_render');
        return (_b = getVueDocument(fileName)) === null || _b === void 0 ? void 0 : _b.getDocument('_render');
    }
    function getTextSpan(document, span) {
        if (isRenderFunctionDocument(document)) {
            const result = document.getOriginalOffsetAt(span.start);
            if (result)
                return { start: result.offset, length: result.length };
        }
        return span;
    }
    return {
        findNodeAtPosition,
        findTemplateElementNodeAt,
        findTemplateNodeAt,
        findTemplateNodeFor,
        findTemplateNodesIn,
        getComponentInfo,
        getDocument,
        getDocumentAt,
        getRenderDoc,
        getTextSpan,
        getVueDocument,
        isRenderFunctionDocument,
        isRenderFunctionFileName,
    };
}
function isNotNull(value) {
    return value != null;
}
function computeIdentifierReplacement(source, identifer) {
    const RE = new RegExp(`\\b${identifer}\\b`);
    const match = RE.exec(source);
    if (!match)
        return { prefixText: source, suffixText: '' };
    return { prefixText: source.substr(0, match.index), suffixText: source.substr(match.index + match[0].length) };
}
function getPaddingLength(source, offset = 0) {
    source = source.substr(offset);
    const match = /^[\s\r\n]+/m.exec(source);
    return match ? match[0].length : 0;
}
function getFilenameForNewComponent(context, directory) {
    let fileName = Path.join(directory, 'Component.vue');
    let index = 0;
    if (context.serviceHost.directoryExists(directory)) {
        const files = new Set(context.serviceHost.readDirectory(directory, ['.vue'], [], [], 1));
        while (files.has(fileName)) {
            fileName = Path.join(directory, `Component${++index}.vue`);
        }
    }
    return fileName;
}
function getComponentName(fileName) {
    return Path.basename(fileName).replace(/\.vue$/, '');
}
function indent(code) {
    return '  ' + code.split('\n').join('\n  ');
}

const noop = {
    getRenameInfo(fileName, position, options) {
        return {
            canRename: false,
            localizedErrorMessage: 'Not implemented.',
        };
    },
    findRenameLocations(fileName, position, findInStrings, findInComments) {
        return undefined;
    },
    getApplicableRefactors(fileName, positionOrRange, preferences) {
        return [];
    },
    getBraceMatchingAtPosition(fileName, position) {
        return [];
    },
    getBreakpointStatementAtPosition(fileName, position) {
        return undefined;
    },
    getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
        return [];
    },
    getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
        return {
            changes: [],
        };
    },
    getCompilerOptionsDiagnostics() {
        return [];
    },
    getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
        return undefined;
    },
    getCompletionEntrySymbol(fileName, position, name, source) {
        return undefined;
    },
    getCompletionsAtPosition(fileName, position, options) {
        return undefined;
    },
    getDefinitionAndBoundSpan(fileName, position) {
        return undefined;
    },
    getDefinitionAtPosition(fileName, position) {
        return undefined;
    },
    getDocCommentTemplateAtPosition(fileName, position) {
        return undefined;
    },
    getDocumentHighlights(fileName, position, filesToSearch) {
        return undefined;
    },
    getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
        return [];
    },
    getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
        return undefined;
    },
    getEmitOutput(fileName) {
        return { outputFiles: [], emitSkipped: true };
    },
    getEncodedSemanticClassifications(fileName, span) {
        return {
            spans: [],
            endOfLineState: 0,
        };
    },
    getEncodedSyntacticClassifications(fileName, span) {
        return {
            spans: [],
            endOfLineState: 0,
        };
    },
    getFormattingEditsAfterKeystroke(fileName, position, key, options) {
        return [];
    },
    getFormattingEditsForDocument(fileName, options) {
        return [];
    },
    getFormattingEditsForRange(fileName, start, end, options) {
        return [];
    },
    getImplementationAtPosition(fileName, position) {
        return undefined;
    },
    getIndentationAtPosition(fileName, position, options) {
        return 0;
    },
    getJsxClosingTagAtPosition(fileName, position) {
        return undefined;
    },
    getNameOrDottedNameSpan(fileName, startPos, endPos) {
        return undefined;
    },
    getNavigateToItems(searchValue) {
        return [];
    },
    getNavigationBarItems(fileName) {
        return [];
    },
    getNavigationTree(fileName) {
        return {
            text: '',
            kind: 'module',
            kindModifiers: 'component',
            spans: [],
            nameSpan: undefined,
        };
    },
    getOccurrencesAtPosition(fileName, position) {
        return undefined;
    },
    getOutliningSpans(fileName) {
        return [];
    },
    getQuickInfoAtPosition(fileName, position) {
        return undefined;
    },
    getReferencesAtPosition(fileName, position) {
        return undefined;
    },
    getSemanticClassifications(fileName, span) {
        return [];
    },
    getSemanticDiagnostics(fileName) {
        return [];
    },
    getSignatureHelpItems(fileName, position, options) {
        return undefined;
    },
    getSmartSelectionRange(fileName, position) {
        return {
            textSpan: { start: position, length: 0 },
        };
    },
    getSpanOfEnclosingComment(fileName, position, onlyMultiLine) {
        return undefined;
    },
    getSuggestionDiagnostics(fileName) {
        return [];
    },
    getSyntacticClassifications(fileName, span) {
        return [];
    },
    getSyntacticDiagnostics(fileName) {
        return [];
    },
    getTodoComments(fileName, descriptors) {
        return [];
    },
    getTypeDefinitionAtPosition(fileName, position) {
        return [];
    },
    findReferences(fileName, position) {
        return undefined;
    },
    prepareCallHierarchy(fileName, position) {
        return undefined;
    },
    provideCallHierarchyIncomingCalls(fileName, position) {
        return [];
    },
    provideCallHierarchyOutgoingCalls(fileName, position) {
        return [];
    },
    isValidBraceCompletionAtPosition(fileName, position, openingBrace) {
        return false;
    },
    organizeImports(scope, formatOptions, preferences) {
        return [];
    },
    getProgram() {
        throw new Error('Not Implemented');
    },
    cleanupSemanticCache() {
        throw new Error('Not Implemented');
    },
    dispose() {
        throw new Error('Not Implemented');
    },
    applyCodeActionCommand() {
        return Promise.reject('Not Implemented');
    },
};

function isNode(node) {
    return typeof node === 'object' && node != null && 'type' in node;
}
function isSimpleExpressionNode(node) {
    return isNode(node) && node.type === 4;
}

const RenameExpression = {
    version: '*',
    canRename(config, fileName, position, preferences) {
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (isSimpleExpressionNode(node)) {
            const mappedPosition = document.getGeneratedOffsetAt(position);
            if (mappedPosition.offset) {
                const result = config.service.getRenameInfo(fileName, position, preferences);
                if (result.canRename) {
                    if (result.displayName === '$event') {
                        return {
                            canRename: false,
                            localizedErrorMessage: '$event is builtin variable, it cannot be renamed.',
                        };
                    }
                    result.triggerSpan = config.helpers.getTextSpan(document, result.triggerSpan);
                }
                return result;
            }
        }
    },
    applyRename(config, fileName, position, findInStrings, findInComments) {
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (isSimpleExpressionNode(node)) {
            const mappedPosition = document.getGeneratedOffsetAt(position);
            if (mappedPosition.offset) {
                return config.service.findRenameLocations(fileName, position, findInStrings, findInComments).slice();
            }
        }
    },
    applyFileRename() {
        return undefined;
    },
};

const RenameComponentTag = {
    version: '*',
    canRename(config, fileName, position) {
        var _a;
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (templateAstTypes.isComponentNode(node) && isPositionInTagName(position, node)) {
            const info = config.helpers.getComponentInfo(document.container);
            const name = node.tag;
            const component = info.components.find((component) => component.name === name);
            let fileToRename = undefined;
            if (component &&
                !component.source.exportName &&
                component.source.moduleName.endsWith('.vue') &&
                getComponentName(component.source.moduleName) === name) {
                fileToRename = component.source.moduleName = (_a = config.context.moduleResolutionHistory
                    .get(fileName)) === null || _a === void 0 ? void 0 : _a.get(component.source.moduleName);
            }
            if (!component) {
                return {
                    canRename: false,
                    localizedErrorMessage: 'Global component (with unknown source file) cannot be renamed.',
                };
            }
            return {
                canRename: true,
                displayName: node.tag,
                fullDisplayName: node.tag,
                kind: config.context.typescript.ScriptElementKind.unknown,
                kindModifiers: 'componentTagName',
                fileToRename,
                triggerSpan: {
                    start: node.loc.start.offset + 1,
                    length: node.tag.length,
                },
            };
        }
    },
    applyRename(config, fileName, position, findInStrings, findInComments) {
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (templateAstTypes.isComponentNode(node) && isPositionInTagName(position, node)) {
            const info = config.helpers.getComponentInfo(document.container);
            const name = node.tag;
            const component = info.components.find((component) => component.name === name);
            const renameLocations = [];
            if (component) {
                if (component.kind === 'script') {
                    const { prefixText } = computeIdentifierReplacement(component.loc.source, component.source.localName);
                    const locations = config.service.findRenameLocations(document.container.getDocumentFileName('script'), component.loc.start.offset + prefixText.length, findInStrings, findInComments);
                    const start = component.source.loc.start.offset;
                    const end = component.source.loc.end.offset;
                    if (locations) {
                        const vueFileName = document.container.fsPath;
                        locations.forEach((location) => {
                            if (vueVirtualTextdocument.isVirtualFile(location.fileName) &&
                                vueVirtualTextdocument.getContainingFile(location.fileName) === vueFileName &&
                                !location.fileName.endsWith('_render.tsx') &&
                                (location.textSpan.start < start || end < location.textSpan.start)) {
                                // rename locations except in render file and component registration location.
                                renameLocations.push(location);
                            }
                        });
                    }
                    const importReplacement = computeIdentifierReplacement(component.source.loc.source, component.source.localName);
                    // add alias to named import.
                    if (component.source.exportName && !importReplacement.prefixText.trim().endsWith(' as')) {
                        importReplacement.prefixText = importReplacement.prefixText + component.source.exportName + ' as ';
                    }
                    renameLocations.unshift({
                        fileName: document.container.fsPath,
                        textSpan: { start: component.source.loc.start.offset, length: component.source.loc.source.length },
                        ...importReplacement,
                    });
                    // if component is registered with an alias, then update to reflect in template.
                    if (component.source.localName !== component.name) {
                        renameLocations.push({
                            fileName: document.container.fsPath,
                            textSpan: { start: component.loc.start.offset, length: component.loc.source.length },
                            ...computeIdentifierReplacement(component.loc.source, component.name),
                        });
                    }
                }
                templateAstTypes.traverseFast(document.ast, (node) => {
                    if (templateAstTypes.isComponentNode(node)) {
                        if (component.aliases.includes(node.tag)) {
                            renameLocations.push({
                                fileName: document.container.fsPath,
                                textSpan: { start: node.loc.start.offset + 1, length: node.tag.length },
                            });
                            if (!node.isSelfClosing) {
                                renameLocations.push({
                                    fileName: document.container.fsPath,
                                    textSpan: {
                                        start: node.loc.start.offset + node.loc.source.lastIndexOf('</' + node.tag) + 2,
                                        length: node.tag.length,
                                    },
                                });
                            }
                        }
                    }
                });
            }
        }
        return undefined;
    },
    applyFileRename(config, containerFile, oldFileName, newFileName, options, preferences) {
        const fileTextChanges = [];
        if (vueVirtualTextdocument.isVueFile(oldFileName)) {
            const document = config.helpers.getRenderDoc(containerFile);
            if (document) {
                const name = getComponentName(oldFileName);
                const newName = getComponentName(newFileName);
                const info = config.helpers.getComponentInfo(document.container);
                const component = info.components.find((component) => component.name === name);
                if (component) {
                    const script = document.container.getDocument('script');
                    if (script) {
                        const position = Math.max(component.loc.end.offset - 2, component.loc.start.offset);
                        const renameInfo = config.service.getRenameInfo(script.fsPath, position);
                        if (renameInfo.canRename) {
                            const edits = config.service.findRenameLocations(script.fsPath, position, false, false);
                            if (edits) {
                                const fileName = document.container.fsPath;
                                fileTextChanges.push({
                                    fileName: fileName,
                                    textChanges: edits
                                        .filter((edit) => vueVirtualTextdocument.isVirtualFile(edit.fileName) && vueVirtualTextdocument.getContainingFile(edit.fileName) === fileName)
                                        .map((edit) => ({ span: edit.textSpan, newText: newName })),
                                });
                                const textChanges = [];
                                // update tags in template
                                templateAstTypes.traverseFast(document.ast, (node) => {
                                    if (templateAstTypes.isComponentNode(node)) {
                                        if (component.aliases.includes(node.tag)) {
                                            textChanges.push({
                                                newText: newName,
                                                span: { start: node.loc.start.offset + 1, length: node.tag.length },
                                            });
                                            if (!node.isSelfClosing) {
                                                textChanges.push({
                                                    newText: newName,
                                                    span: {
                                                        start: node.loc.start.offset + node.loc.source.lastIndexOf('</' + node.tag) + 2,
                                                        length: node.tag.length,
                                                    },
                                                });
                                            }
                                        }
                                    }
                                });
                                fileTextChanges.push({ fileName, textChanges: textChanges });
                            }
                        }
                    }
                }
            }
        }
        if (fileTextChanges.length)
            return fileTextChanges;
    },
};
function isPositionInTagName(position, node) {
    return (
    // In start tag.
    position <= node.loc.start.offset + node.tag.length ||
        // In end tag.
        (!node.isSelfClosing && node.loc.start.offset + node.loc.source.lastIndexOf('</') <= position));
}

const RenameElementTag = {
    version: '*',
    canRename(config, fileName, position) {
        const { node } = config.helpers.findNodeAtPosition(fileName, position);
        if (templateAstTypes.isPlainElementNode(node) && isPositionInTagName$1(position, node)) {
            return {
                canRename: true,
                displayName: node.tag,
                fullDisplayName: node.tag,
                kind: config.context.typescript.ScriptElementKind.unknown,
                kindModifiers: 'elementTagName',
                triggerSpan: {
                    start: node.loc.start.offset + 1,
                    length: node.tag.length,
                },
            };
        }
    },
    applyRename(config, fileName, position) {
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (templateAstTypes.isPlainElementNode(node) && isPositionInTagName$1(position, node)) {
            const locations = [
                {
                    fileName: document.container.fsPath,
                    textSpan: { start: node.loc.start.offset + 1, length: node.tag.length },
                },
            ];
            if (!node.isSelfClosing) {
                locations.push({
                    fileName: document.container.fsPath,
                    textSpan: {
                        start: node.loc.start.offset + node.loc.source.lastIndexOf('</' + node.tag) + 2,
                        length: node.tag.length,
                    },
                });
            }
            return locations;
        }
    },
    applyFileRename() {
        return undefined;
    },
};
function isPositionInTagName$1(position, node) {
    return (
    // In start tag.
    position <= node.loc.start.offset + node.tag.length ||
        // In end tag.
        (!node.isSelfClosing && node.loc.start.offset + node.loc.source.lastIndexOf('</') <= position));
}

const RENAME_PROVIDERS = [RenameExpression, RenameComponentTag, RenameElementTag];

const RefactorExtractComponent = {
    version: '*',
    findRefactors(config, fileName, position, preferences) {
        const document = config.helpers.getRenderDoc(fileName);
        if (document) {
            const nodes = findNodes(config, document, position);
            if (nodes.length) {
                let isExtractSupported = true;
                const modelExpressions = new Set();
                nodes.forEach((node) => templateAstTypes.traverseFast(node, (node) => {
                    if (templateAstTypes.isDirectiveNode(node) && node.name === 'model') {
                        if (templateAstTypes.isSimpleExpressionNode(node.exp)) {
                            modelExpressions.add(node.exp.content.trim());
                        }
                    }
                }));
                if (modelExpressions.size > 1) {
                    const version = parseInt(config.context.getVueVersion(document.container.fsPath));
                    if (version < 3)
                        isExtractSupported = false;
                }
                if (isExtractSupported) {
                    const extractAsComponent = {
                        name: 'component',
                        description: 'Extract to component',
                        inlineable: true,
                        actions: [
                            {
                                name: ':current',
                                description: 'Extract to component in current directory',
                            },
                        ],
                    };
                    const project = config.context.projectService.getDefaultProjectForFile(config.context.typescript.server.toNormalizedPath(document.fsPath), false);
                    if (project) {
                        config.context.config.directories
                            .filter((dir) => dir.kind === 'component')
                            .forEach((config) => {
                            extractAsComponent.actions.push({
                                name: config.name,
                                description: `Extract to component in ${config.name} directory`,
                            });
                        });
                    }
                    if (extractAsComponent.actions.length > 4) {
                        extractAsComponent.inlineable = false;
                    }
                    return [extractAsComponent];
                }
            }
        }
        return [];
    },
    applyRefactor(config, fileName, options, position, refactorName, actionName, preferences) {
        const document = config.helpers.getRenderDoc(fileName);
        if (document && refactorName === 'component') {
            const project = config.context.projectService.getDefaultProjectForFile(config.context.typescript.server.toNormalizedPath(document.fsPath), false);
            const path = actionName === ':current'
                ? Path__default.dirname(fileName)
                : config.context.config.directories.find((directory) => directory.path).path;
            const directoryName = Path__default.isAbsolute(path) ? path : Path__default.resolve(project.getCurrentDirectory(), path);
            const componentFileName = getFilenameForNewComponent(config.context, directoryName);
            const nodes = findNodes(config, document, position);
            const identifiers = Array.from(new Set(nodes.flatMap((node) => node.scope.identifiers.filter((id) => node.scope.getBinding(id) !== node))));
            const template = nodes.map((node) => node.loc.source).join('');
            // TODO: detect types of props.
            const script = [`import { defineComponent } from 'vue'`];
            const components = new Set();
            const models = new Set();
            nodes.forEach((node) => templateAstTypes.traverseFast(node, (node) => {
                if (templateAstTypes.isComponentNode(node))
                    components.add(node.tag);
                else if (templateAstTypes.isDirectiveNode(node) && node.name === 'model') {
                    if (templateAstTypes.isSimpleExpressionNode(node.exp)) {
                        node.exp.scope.globals.forEach((identifier) => {
                            models.add(identifier);
                        });
                    }
                }
            }));
            const info = config.helpers.getComponentInfo(document.container);
            const options = {
                components: `components: {${info.components.reduce((text, { name, source }) => {
                    if (components.has(name)) {
                        script.push(`import ${source.exportName
                            ? `{ ${source.localName === source.exportName ? `` : `${source.exportName} as `}${source.localName} }`
                            : `${source.localName}`} from '${source.moduleName}'`);
                        return (text +
                            `${name === source.localName ? `` : `${/^[a-z0-9]$/i.test(name) ? `'${name}'` : name}: `}${source.localName}, `);
                    }
                    return text;
                }, '')}},`,
                props: `props: [${identifiers.map((identifier) => `'${identifier}'`).join(', ')}],`,
            };
            script.push(`export default defineComponent({\n  ${options.props}\n  ${options.components}\n})`);
            const code = `<script>\n${script.join('\n')}\n</script>\n\n<template>\n${formatTemplate(template)}\n</template>\n`;
            const { name, changes, renameLocation } = getImportEditForComponent(document.container, info, componentFileName);
            changes.push({
                newText: `<${name} ${identifiers
                    .map((identifer) => `${models.has(identifer) ? 'v-model' : ''}:${identifer}="${identifer}"`)
                    .join(' ')}/>`,
                span: {
                    start: nodes[0].loc.start.offset,
                    length: nodes.reduce((len, node) => len + node.loc.source.length, 0),
                },
            });
            config.context.createVueDocument(componentFileName, code);
            return {
                renameFilename: document.container.fsPath,
                renameLocation,
                edits: [
                    {
                        isNewFile: true,
                        fileName: componentFileName,
                        textChanges: [
                            {
                                span: {
                                    start: 0,
                                    length: 0,
                                },
                                newText: code,
                            },
                        ],
                    },
                    {
                        fileName: document.container.fsPath,
                        textChanges: changes,
                    },
                ],
            };
        }
        return undefined;
    },
};
function formatTemplate(template) {
    const index = template.indexOf('\n');
    if (index >= 0) {
        const firstLine = template.substr(0, index + 1);
        const otherLines = template.substr(index + 1);
        return indent(firstLine + deIndent(otherLines));
    }
    return indent(template);
}
function findNodes(config, document, position) {
    return typeof position === 'number'
        ? config.helpers.findTemplateNodesIn(document.ast, position, position)
        : config.helpers.findTemplateNodesIn(document.ast, position.pos, position.end);
}
function getImportEditForComponent(document, info, fileName) {
    const changes = [];
    const name = getComponentName(fileName);
    const { script } = document.descriptor;
    const relativeFileName = `./${Path__default.relative(Path__default.dirname(document.fsPath), fileName)}`;
    let renameLocation = 0;
    if (script && !script.setup) {
        const newText = `\nimport ${name} from '${relativeFileName}';`;
        renameLocation = script.loc.start.offset + newText.indexOf(relativeFileName);
        changes.push({
            newText,
            span: { start: script.loc.start.offset, length: 0 },
        });
        const options = info.options;
        const components = options.properties['components'];
        if (components) {
            const start = components.loc.start.offset + 1;
            const padding = getPaddingLength(components.loc.source, 1);
            changes.push({
                newText: components.loc.source.substr(1, padding) + `${name},`,
                span: { start, length: 0 },
            });
        }
        else if (options) {
            const start = options.loc.start.offset + 1;
            const padding = getPaddingLength(options.loc.source, 1);
            changes.push({
                newText: components.loc.source.substr(1, padding) + `\ncomponents: { ${name} },`,
                span: { start, length: 0 },
            });
        }
    }
    else {
        const newText = `<component src="./${Path__default.relative(Path__default.dirname(document.fsPath), fileName)}" />\n\n`;
        renameLocation = newText.indexOf(relativeFileName);
        changes.push({
            newText,
            span: { start: 0, length: 0 },
        });
    }
    return { name, changes, renameLocation };
}

const REFACTOR_PROVIDERS = [RefactorExtractComponent];

function createTemplateLanguageServer(config) {
    const { helpers: h, service, context } = config;
    const cache = new QuickLRU({ maxSize: 1000 });
    return {
        ...noop,
        getQuickInfoAtPosition(fileName, position) {
            const document = h.getRenderDoc(fileName);
            if (!document)
                return;
            const loc = document.getGeneratedOffsetAt(position);
            if (!loc)
                return;
            const result = service.getQuickInfoAtPosition(fileName, loc.offset);
            if (result) {
                const textSpan = h.getTextSpan(document, result.textSpan);
                if (textSpan) {
                    result.textSpan = textSpan;
                    return result;
                }
            }
        },
        getSemanticDiagnostics(fileName) {
            const document = h.getRenderDoc(fileName);
            if (!document)
                return [];
            const key = `getSemanticDiagnostics::${document.container.version}::${fileName}`;
            if (cache.has(key))
                return cache.get(key);
            const diagnostics = service
                .getSemanticDiagnostics(fileName)
                .map((diagnostic) => {
                var _a;
                if (Number.isInteger(diagnostic.start)) {
                    const position = document.findExpression(diagnostic.start, diagnostic.length || 1);
                    context.log(`TRY RENDER => ${(_a = diagnostic.file) === null || _a === void 0 ? void 0 : _a.fileName} --- ${diagnostic.messageText} -- ${JSON.stringify(position)}`);
                    if (!position)
                        return;
                    diagnostic.start = position.offset;
                    diagnostic.length = position.length;
                }
                return diagnostic;
            })
                .filter(isNotNull);
            cache.set(key, diagnostics);
            return diagnostics;
        },
        getSuggestionDiagnostics(fileName) {
            const document = h.getRenderDoc(fileName);
            if (!document)
                return [];
            const key = `getSuggestionDiagnostics::${document.container.version}::${fileName}`;
            if (cache.has(key))
                return cache.get(key);
            const diagnostics = service
                .getSuggestionDiagnostics(fileName)
                .map((diagnostic) => {
                if (Number.isInteger(diagnostic.start)) {
                    const position = document.findExpression(diagnostic.start, diagnostic.length || 1);
                    if (!position)
                        return;
                    diagnostic.start = position.offset;
                    diagnostic.length = position.length;
                }
                return diagnostic;
            })
                .filter(isNotNull);
            cache.set(key, diagnostics);
            return diagnostics;
        },
        getSyntacticDiagnostics(fileName) {
            const document = h.getRenderDoc(fileName);
            if (!document)
                return [];
            const key = `getSyntacticDiagnostics::${document.container.version}::${fileName}`;
            if (cache.has(key))
                return cache.get(key);
            const diagnostics = service
                .getSyntacticDiagnostics(fileName)
                .map((diagnostic) => {
                if (Number.isInteger(diagnostic.start)) {
                    const position = document.findExpression(diagnostic.start, diagnostic.length || 1);
                    if (!position)
                        return;
                    diagnostic.start = position.offset;
                    diagnostic.length = position.length;
                }
                return diagnostic;
            })
                .filter(isNotNull);
            cache.set(key, diagnostics);
            return diagnostics;
        },
        getRenameInfo(fileName, position, preferences) {
            for (const provider of RENAME_PROVIDERS) {
                const result = provider.canRename(config, fileName, position, preferences);
                if (result)
                    return result;
            }
            return {
                canRename: false,
                localizedErrorMessage: '',
            };
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            for (const provider of RENAME_PROVIDERS) {
                const result = provider.applyRename(config, fileName, position, findInStrings, findInComments);
                if (result)
                    return result;
            }
            return [];
        },
        getEditsForFileRenameIn(fileName, oldFilePath, newFilePath) {
            for (const provider of RENAME_PROVIDERS) {
                const result = provider.applyFileRename(config, fileName, oldFilePath, newFilePath, {}, {});
                if (result)
                    return result;
            }
            return [];
        },
        getApplicableRefactors(fileName, position, preferences) {
            const refactors = [];
            for (const provider of REFACTOR_PROVIDERS) {
                const result = provider.findRefactors(config, fileName, position, preferences);
                if (result)
                    refactors.push(...result);
            }
            return refactors;
        },
        getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
            for (const provider of REFACTOR_PROVIDERS) {
                const result = provider.applyRefactor(config, fileName, formatOptions, positionOrRange, refactorName, actionName, preferences);
                if (result)
                    return result;
            }
        },
    };
}

function createVueLanguageServer(options) {
    const template = createTemplateLanguageServer(options);
    const { helpers: h, service: script, context } = options;
    function isFeatureEnabled(featureName, checkFor = true) {
        const feature = context.config.features[featureName];
        return Array.isArray(feature) ? feature.includes(checkFor) : feature === checkFor;
    }
    function choose(document) {
        return h.isRenderFunctionDocument(document) ? template : script;
    }
    return {
        ...noop,
        organizeImports(scope, formatOptions, preferences) {
            if (!isFeatureEnabled('organizeImports'))
                return [];
            const document = h.getVueDocument(scope.fileName);
            if (document) {
                const virtual = document.getDocument('scriptSetup') || document.getDocument('script');
                if (virtual) {
                    return script.organizeImports({ ...scope, fileName: virtual.fsPath }, formatOptions, preferences);
                }
            }
            return [];
        },
        getSemanticDiagnostics(fileName) {
            if (!isFeatureEnabled('diagnostics', 'semantic'))
                return [];
            const document = h.getVueDocument(fileName);
            const diagnostics = [];
            if (document) {
                ['script', '_render'].forEach((selector) => {
                    const virtual = document.getDocument(selector);
                    if (virtual) {
                        const results = choose(virtual).getSemanticDiagnostics(virtual.fsPath);
                        diagnostics.push(...results);
                    }
                });
            }
            return diagnostics;
        },
        getQuickInfoAtPosition(fileName, position) {
            if (!isFeatureEnabled('quickInfo'))
                return;
            // TODO: Provide better quick info for components and props.
            const document = h.getDocumentAt(fileName, position);
            if (document)
                return choose(document).getQuickInfoAtPosition(document.fsPath, position);
        },
        getSuggestionDiagnostics(fileName) {
            if (!isFeatureEnabled('diagnostics', 'suggestion'))
                return [];
            const document = h.getVueDocument(fileName);
            const diagnostics = [];
            if (document) {
                ['script', '_render'].forEach((selector) => {
                    const virtual = document.getDocument(selector);
                    if (virtual)
                        diagnostics.push(...choose(virtual).getSuggestionDiagnostics(virtual.fsPath));
                });
            }
            return diagnostics;
        },
        getSyntacticDiagnostics(fileName) {
            if (!isFeatureEnabled('diagnostics', 'syntactic'))
                return [];
            const document = h.getVueDocument(fileName);
            const diagnostics = [];
            if (document) {
                ['script', '_render'].forEach((selector) => {
                    const virtual = document.getDocument(selector);
                    if (virtual)
                        diagnostics.push(...choose(virtual).getSyntacticDiagnostics(virtual.fsPath));
                });
            }
            return diagnostics;
        },
        getRenameInfo(fileName, position, options) {
            if (!isFeatureEnabled('rename'))
                return { canRename: false, localizedErrorMessage: 'Rename feature disabled.' };
            const document = h.getDocumentAt(fileName, position);
            if (!document) {
                return {
                    canRename: false,
                    localizedErrorMessage: 'Cannot find file',
                };
            }
            return choose(document).getRenameInfo(document.fsPath, position, options);
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            if (!isFeatureEnabled('rename'))
                return [];
            const document = h.getVueDocument(fileName);
            if (!document)
                return;
            const block = document.blockAt(position);
            if (!block)
                return;
            const result = [];
            if (block.type === 'template' || block.type === 'script') {
                const fromTemplate = template.findRenameLocations(document.getDocumentFileName('_render'), position, findInStrings, findInComments);
                if (fromTemplate)
                    result.push(...fromTemplate);
            }
            if (block.type === 'script') {
                const fromScript = template.findRenameLocations(document.getDocumentFileName('script'), position, findInStrings, findInComments);
                if (fromScript)
                    result.push(...fromScript);
            }
            result.sort((a, b) => a.textSpan.start - b.textSpan.start);
            return result;
        },
        getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
            if (!isFeatureEnabled('rename'))
                return [];
            const document = h.getVueDocument(oldFilePath);
            const result = [];
            if (document) {
                const doc = document.getDocument('_module');
                result.push(...script.getEditsForFileRename(doc.fsPath, doc.fsPath.replace(oldFilePath, newFilePath), formatOptions, preferences));
                const visited = new Set();
                result.forEach((item) => {
                    var _a;
                    if (vueVirtualTextdocument.isVirtualFile(item.fileName) || vueVirtualTextdocument.isVueFile(item.fileName)) {
                        const render = (_a = h.getVueDocument(item.fileName)) === null || _a === void 0 ? void 0 : _a.getDocument('_render');
                        if (render && !visited.has(render.fsPath)) {
                            visited.add(render.fsPath);
                            result.push(...template.getEditsForFileRenameIn(render.fsPath, oldFilePath, newFilePath));
                        }
                    }
                });
            }
            return result;
        },
        getApplicableRefactors(fileName, positionOrRange, preferences) {
            if (!isFeatureEnabled('refactor'))
                return [];
            const document = h.getDocumentAt(fileName, vueVirtualTextdocument.isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos);
            const document2 = h.getDocumentAt(fileName, vueVirtualTextdocument.isNumber(positionOrRange) ? positionOrRange : positionOrRange.end);
            if (document && document === document2) {
                return choose(document).getApplicableRefactors(document.fsPath, positionOrRange, preferences);
            }
            return [];
        },
        getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
            if (!isFeatureEnabled('refactor'))
                return;
            const document = h.getDocumentAt(fileName, vueVirtualTextdocument.isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos);
            const document2 = h.getDocumentAt(fileName, vueVirtualTextdocument.isNumber(positionOrRange) ? positionOrRange : positionOrRange.end);
            if (document && document === document2) {
                return choose(document).getEditsForRefactor(document.fsPath, formatOptions, positionOrRange, refactorName, actionName, preferences);
            }
        },
    };
}

const VUE_LANGUAGE_SERVER = Symbol('Vue Language Server');
class RoutingLanguageServer {
    constructor(context) {
        this.context = context;
    }
    decorate(languageService) {
        if (VUE_LANGUAGE_SERVER in languageService) {
            return languageService;
        }
        const proxy = createLanguageServiceRouter({
            context: this.context,
            service: languageService,
            helpers: createServerHelper(this.context),
        });
        // @ts-ignore
        proxy[VUE_LANGUAGE_SERVER] = true;
        return proxy;
    }
}
function createLanguageServiceRouter(options) {
    const vue = createVueLanguageServer(options);
    const ts = options.service;
    function choose(fileName) {
        return vueVirtualTextdocument.isVueFile(fileName) ? vue : ts;
    }
    const methods = [
        'getRenameInfo',
        'findRenameLocations',
        'getApplicableRefactors',
        'getBraceMatchingAtPosition',
        'getBreakpointStatementAtPosition',
        'getCodeFixesAtPosition',
        'getCompletionEntryDetails',
        'getCompletionEntrySymbol',
        'getCompletionsAtPosition',
        'getDefinitionAndBoundSpan',
        'getDefinitionAtPosition',
        'getDocCommentTemplateAtPosition',
        'getDocumentHighlights',
        'getEditsForFileRename',
        'getEditsForRefactor',
        'getEmitOutput',
        'getEncodedSemanticClassifications',
        'getEncodedSyntacticClassifications',
        'getFormattingEditsAfterKeystroke',
        'getFormattingEditsForDocument',
        'getFormattingEditsForRange',
        'getImplementationAtPosition',
        'getIndentationAtPosition',
        'getJsxClosingTagAtPosition',
        'getNameOrDottedNameSpan',
        'getNavigateToItems',
        'getNavigationBarItems',
        'getNavigationTree',
        'getOccurrencesAtPosition',
        'getOutliningSpans',
        'getQuickInfoAtPosition',
        'getReferencesAtPosition',
        'getSemanticClassifications',
        'getSemanticClassifications',
        'getSignatureHelpItems',
        'getSmartSelectionRange',
        'getSpanOfEnclosingComment',
        'getSuggestionDiagnostics',
        'getSyntacticDiagnostics',
        'getTodoComments',
        'getTypeDefinitionAtPosition',
        'findReferences',
        'prepareCallHierarchy',
        'provideCallHierarchyIncomingCalls',
        'provideCallHierarchyOutgoingCalls',
        'isValidBraceCompletionAtPosition',
    ];
    function getTextSpan(document, span) {
        if (options.helpers.isRenderFunctionDocument(document)) {
            const result = document.getOriginalOffsetAt(span.start);
            if (result)
                return { start: result.offset, length: result.length };
            return null;
        }
        return span;
    }
    const VIRTUAL_FILE_SUFFIX_RE = new RegExp(`(?<=\.vue)${vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR}([A-Za-z_][A-Za-z0-9_-]*)(\\.[jt]sx?)?`, 'g');
    function getComponentName(fileName) {
        if (!fileName || !vueVirtualTextdocument.isVueFile(fileName))
            return;
        const baseName = Path.basename(fileName);
        return baseName.substr(0, baseName.length - 4);
    }
    const COMPONENT_TYPE_RE = /'ComponentPublicInstance<.*?ComponentOptionsBase<...>>'/g;
    const REPLACE = {
        virtualFile: (messageText) => messageText.replace(VIRTUAL_FILE_SUFFIX_RE, ''),
        componentType: (fileName, messageText) => {
            const component = getComponentName(fileName);
            return component ? messageText.replace(COMPONENT_TYPE_RE, `vue component '${component}'`) : messageText;
        },
    };
    function applyReplacements(fileName, messageText) {
        if (typeof messageText === 'string') {
            return REPLACE.componentType(fileName, REPLACE.virtualFile(messageText));
        }
        else {
            messageText.messageText = applyReplacements(fileName, messageText.messageText);
            return messageText;
        }
    }
    function isVirtualSourceFile(file) {
        return !!file && vueVirtualTextdocument.isVirtualFile(file.fileName);
    }
    const proxy = {
        organizeImports(scope, formatOptions, preferences) {
            return choose(scope.fileName)
                .organizeImports(scope, formatOptions, preferences)
                .map((change) => {
                if (vueVirtualTextdocument.isVirtualFile(change.fileName))
                    change.fileName = vueVirtualTextdocument.getContainingFile(change.fileName);
                return change;
            });
        },
        getQuickInfoAtPosition(fileName, position) {
            const info = choose(fileName).getQuickInfoAtPosition(fileName, position);
            if (info === null || info === void 0 ? void 0 : info.displayParts) {
                info.displayParts = info.displayParts
                    .map((part) => {
                    part.text = applyReplacements(fileName, part.text);
                    return part;
                })
                    .filter(isNotNull);
            }
            if (info === null || info === void 0 ? void 0 : info.documentation) {
                info.documentation = info.documentation
                    .map((part) => {
                    part.text = applyReplacements(fileName, part.text);
                    return part;
                })
                    .filter(isNotNull);
            }
            return info;
        },
        getSemanticDiagnostics(fileName) {
            const diagnostics = choose(fileName).getSemanticDiagnostics(fileName);
            const program = options.service.getProgram();
            return diagnostics
                .map((diagnostic) => {
                if (isVirtualSourceFile(diagnostic.file)) {
                    diagnostic.file = { ...diagnostic.file, fileName: vueVirtualTextdocument.getContainingFile(diagnostic.file.fileName) };
                }
                diagnostic.messageText = applyReplacements(fileName, diagnostic.messageText);
                if (diagnostic.relatedInformation) {
                    diagnostic.relatedInformation = diagnostic.relatedInformation.map((info) => {
                        info.messageText = applyReplacements(fileName, info.messageText);
                        if (isVirtualSourceFile(info.file)) {
                            info.file = program === null || program === void 0 ? void 0 : program.getSourceFile(vueVirtualTextdocument.getContainingFile(info.file.fileName));
                        }
                        return info;
                    });
                }
                return diagnostic;
            })
                .filter(isNotNull);
        },
        getSyntacticDiagnostics(fileName) {
            const diagnostics = choose(fileName).getSyntacticDiagnostics(fileName);
            const program = options.service.getProgram();
            return diagnostics
                .map((diagnostic) => {
                if (isVirtualSourceFile(diagnostic.file)) {
                    diagnostic.file = { ...diagnostic.file, fileName: vueVirtualTextdocument.getContainingFile(diagnostic.file.fileName) };
                }
                diagnostic.messageText = applyReplacements(fileName, diagnostic.messageText);
                if (diagnostic.relatedInformation) {
                    diagnostic.relatedInformation = diagnostic.relatedInformation.map((info) => {
                        info.messageText = applyReplacements(fileName, info.messageText);
                        if (isVirtualSourceFile(info.file)) {
                            info.file = program === null || program === void 0 ? void 0 : program.getSourceFile(vueVirtualTextdocument.getContainingFile(info.file.fileName));
                        }
                        return info;
                    });
                }
                return diagnostic;
            })
                .filter(isNotNull);
        },
        getSuggestionDiagnostics(fileName) {
            const diagnostics = choose(fileName).getSuggestionDiagnostics(fileName);
            const program = options.service.getProgram();
            return diagnostics
                .map((diagnostic) => {
                if (isVirtualSourceFile(diagnostic.file)) {
                    diagnostic.file = { ...diagnostic.file, fileName: vueVirtualTextdocument.getContainingFile(diagnostic.file.fileName) };
                }
                diagnostic.messageText = applyReplacements(fileName, diagnostic.messageText);
                if (diagnostic.relatedInformation) {
                    diagnostic.relatedInformation = diagnostic.relatedInformation.map((info) => {
                        info.messageText = applyReplacements(fileName, info.messageText);
                        if (isVirtualSourceFile(info.file)) {
                            info.file = { ...info.file, fileName: vueVirtualTextdocument.getContainingFile(info.file.fileName) };
                        }
                        return info;
                    });
                }
                return diagnostic;
            })
                .filter(isNotNull);
        },
        getRenameInfo(fileName, position, options) {
            const result = choose(fileName).getRenameInfo(fileName, position, options);
            if (result.canRename) {
                if (result.fileToRename && vueVirtualTextdocument.isVirtualFile(result.fileToRename)) {
                    result.fileToRename = vueVirtualTextdocument.getContainingFile(result.fileToRename);
                }
                result.displayName = REPLACE.virtualFile(result.displayName);
                result.fullDisplayName = REPLACE.virtualFile(result.fullDisplayName);
            }
            return result;
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            var _a;
            return (_a = choose(fileName)
                .findRenameLocations(fileName, position, findInStrings, findInComments)) === null || _a === void 0 ? void 0 : _a.map((item) => {
                options.context.log('try.findRenameLocations ' + JSON.stringify(item));
                if (vueVirtualTextdocument.isVirtualFile(item.fileName)) {
                    item.originalContextSpan = item.contextSpan;
                    item.originalTextSpan = item.textSpan;
                    item.originalFileName = item.fileName;
                    const virtual = options.helpers.getDocument(item.fileName);
                    item.fileName = virtual.container.fsPath;
                    const textSpan = getTextSpan(virtual, item.textSpan);
                    if (!textSpan)
                        return;
                    item.textSpan = textSpan;
                    if (item.contextSpan) {
                        const contextSpan = getTextSpan(virtual, item.contextSpan);
                        if (!contextSpan)
                            return;
                        item.contextSpan = contextSpan;
                    }
                }
                options.context.log('ok.findRenameLocations ' + JSON.stringify(item));
                return item;
            }).filter(isNotNull);
        },
        getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
            const suffix = '.vue' + vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR + '_module';
            return choose(oldFilePath)
                .getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences)
                .map((edit) => {
                if (vueVirtualTextdocument.isVirtualFile(edit.fileName)) {
                    const selector = vueVirtualTextdocument.parseVirtualFileName(edit.fileName);
                    if ((selector === null || selector === void 0 ? void 0 : selector.selector.type) !== 'script')
                        return;
                    edit.fileName = vueVirtualTextdocument.getContainingFile(edit.fileName);
                }
                edit.textChanges = edit.textChanges
                    .map((change) => {
                    const index = change.newText.indexOf(suffix);
                    if (index >= 0) {
                        change.newText = change.newText.substr(0, index + 4) + change.newText.substr(index + suffix.length);
                    }
                    return change;
                })
                    .filter(isNotNull);
                if (!edit.textChanges.length)
                    return;
                return edit;
            })
                .filter(isNotNull);
        },
        getApplicableRefactors(fileName, positionOrRange, preferences) {
            return choose(fileName).getApplicableRefactors(fileName, positionOrRange, preferences);
        },
    };
    methods.forEach((name) => {
        if (!(name in proxy)) {
            // @ts-ignore
            proxy[name] = function (fileName) {
                const service = choose(fileName);
                // @ts-ignore
                return service[name].apply(service, arguments);
            };
        }
    });
    return proxy;
}

let context;
let server;
function init({ typescript }) {
    if (!context) {
        context = new PluginContext(typescript);
        server = new RoutingLanguageServer(context);
    }
    return {
        create(info) {
            const version = findVueVersionIn(info.project.getCurrentDirectory());
            if (!isVue3(version)) {
                return info.languageService;
            }
            else {
                context.load(info);
                return server.decorate(info.languageService);
            }
        },
        onConfigurationChanged(config) {
            if (context) {
                context.setConfig(config);
            }
        },
    };
}
function isVue3(version) {
    if (!version)
        return true; // assume Vue 3 by default.
    return version.startsWith('3.');
}
function findVueVersionIn(directory) {
    var _a, _b;
    const packageFile = findNearestPackageFile(directory);
    if (packageFile) {
        const vuePackageFile = Path__default.resolve(Path__default.dirname(packageFile), 'node_modules', 'vue', 'package.json');
        if (FS__default.existsSync(vuePackageFile)) {
            return require(vuePackageFile).version;
        }
        // dependencies not installed.
        const pkg = require(packageFile);
        const version = ((_a = pkg.dependencies) === null || _a === void 0 ? void 0 : _a.vue) || ((_b = pkg.devDependencies) === null || _b === void 0 ? void 0 : _b.vue);
        if (!version)
            return;
        const justNumbers = version.replace(/[^0-9.]/, '');
        if (justNumbers.startsWith('3.'))
            return justNumbers;
        if (version === 'next')
            return '3.0.0';
        return '2.x';
    }
}
function findNearestPackageFile(directory) {
    if (!directory)
        return;
    const packageFile = Path__default.resolve(directory, 'package.json');
    if (FS__default.existsSync(packageFile))
        return packageFile;
    if (directory === '/')
        return;
    return findNearestPackageFile(Path__default.dirname(directory));
}

module.exports = init;
//# sourceMappingURL=typescript-plugin-vue.js.map
